"""Module for handling authentication."""


from datetime import datetime, timedelta
from typing import Any, Dict

import jwt
import requests
from jwt import PyJWKClient, PyJWTError

from peachpayments_partner import config
from peachpayments_partner.exceptions import AuthError
from peachpayments_partner.jwt_formatter import TokenFormatter

jwks_client = PyJWKClient(config.AUTH0_JWKS_URL)


def validate_token(value: str) -> str:
    """Validates token.

    Args:
       value (str): JWT token string

    Returns:
        a validated token

    Raises:
        AuthError if invalid token is provided
    """
    if not isinstance(value, str):
        raise AuthError("Invalid token type, expecting token of type str")
    parts = value.split()

    if len(parts) != 2:
        raise AuthError("Invalid Authorization token, must be Bearer <token>")

    if len(parts) == 2 and parts[0].lower() != "bearer":
        raise AuthError("Authorization header must start with Bearer")

    is_valid_token = TokenFormatter.is_valid_jwt(parts[1])
    if not is_valid_token:
        raise AuthError("Invalid JWT format provided")
    # Get just the token part of the string
    return parts[1]


def get_key(token: str) -> jwt.PyJWK:
    """Gets key from JWT token.

    Args:
        token (str) : JWT token to be checked for authentication

    Returns:
        a jwt.PyJWK key

    Raises:
        AuthError if key generation fails

    """
    valid_token = validate_token(token)
    try:
        signing_key = jwks_client.get_signing_key_from_jwt(valid_token)
    except PyJWTError as key_error:
        raise AuthError("key generation failed: %s", key_error)
    return signing_key.key


def is_authenticated(token: str, signing_key: str) -> bool:
    """Checks if a key is valid.

    Args:
        token (str) : JWT token to be checked for authentication
        signing_key (str): JWK signing key object

    Returns:
         Boolean value of whether key is valid or not

    Raises:
        AuthError for decode errors.
    """
    valid_token = validate_token(token)
    try:
        jwt.decode(valid_token, signing_key, algorithms=["RS256"], audience=config.AUDIENCE)
    except PyJWTError:
        raise AuthError("Token validation Failed")

    return True


_access_token_cache: Dict[str, Dict[str, Any]] = {}


def get_access_token(client_id: str, client_secret: str) -> str:
    """Retrieve an access token for a client id+secret pair.

    Also caches the token for a period of time to reduce network traffic when creating a token.

    Args:
       client_id (str): Application client id
       client_secret (str): Application client secret

    Returns:
        An access token for use with Peach services

    Raises:
        AuthError if an access token cannot be retrieved.
    """
    key = f"{client_id}:{client_secret}"
    if key in _access_token_cache:
        value = _access_token_cache[key]
        if datetime.utcnow() < value.get("expiration", datetime.utcnow()):
            token: str = value.get("value", "")
            return token
    headers = {
        "Content-Type": "application/json",
        "Accept": "*/*",
    }
    body = {
        "clientId": client_id,
        "clientSecret": client_secret,
    }

    response = requests.post(f"{config.AUTH_SERVICE}oauth/token", json=body, headers=headers)

    if response.ok:
        response_body: dict = response.json()
        expires_in = timedelta(seconds=response_body.get("expires_in", 5) - 5)
        access_token: str = response_body.get("access_token", "")

        _access_token_cache.update(
            {
                key: {
                    "value": access_token,
                    "expiration": datetime.utcnow() + expires_in,
                }
            }
        )

        return access_token

    raise AuthError("Unable to retrieve an access token.")
