"""Provides endpoint field validation."""

from datetime import datetime, timezone
from typing import Optional, Type, Union

from ..result_codes import ResultCode, result_codes
from .schemas import DebitRequestSchema, DebitResponseSchema, RefundRequestSchema, RefundResponseSchema

ParameterErrorType = dict[str, Optional[str]]
ParameterErrorsType = list[ParameterErrorType]
ErrorResultType = dict[str, Union[ParameterErrorsType, Optional[str]]]
ErrorResponseType = dict[str, Union[str, ErrorResultType]]
ValidateRequestReturnType = dict[
    str,
    Union[
        bool,
        list[str],
        ErrorResponseType,
    ],
]
ValidateResponseReturnType = dict[str, Union[bool, list[str]]]


def _validate_request(schema: Type, data: dict) -> ValidateRequestReturnType:
    """Validate Peach request fields.

    Args:
        schema: Schema class defined in .schemas
        data (dict): request fields as defined in spec

    Returns:
        (dict): validated data and response fields if not valid

    The `response` fields are used to return the error message to the client.

    """
    errors = schema().validate(data)
    if not errors:
        return dict(valid=True)

    # TODO modify result_codes to support typing
    result_code = result_codes.INVALID_OR_MISSING_PARAMETER  # type: ignore
    return dict(valid=False, errors=errors, response=format_error_response(result_code, errors, data))


def _validate_response(schema: Type, data: dict) -> ValidateResponseReturnType:
    """Validate endpoint response fields.

    Args:
        schema: Schema class defined in .schemas
        data (dict): response fields as defined in spec

    Returns:
        (dict): validated data
    """
    errors = schema().validate(data)
    if not errors:
        return dict(valid=True)

    return dict(valid=False, errors=errors)


def validate_debit_request(data: dict) -> ValidateRequestReturnType:
    """Validate Debit request data.

    Args:
        data (dict): debit request fields as defined in spec

    Returns:
        (dict): validated data
    """
    return _validate_request(DebitRequestSchema, data)


def validate_debit_response(data: dict) -> ValidateResponseReturnType:
    """Validate Debit response data.

    Args:
        data (dict): debit response fields as defined in spec

    Returns:
        (dict): validated data
    """
    return _validate_response(DebitResponseSchema, data)


def validate_refund_request(data: dict) -> ValidateRequestReturnType:
    """Validate Refund request data.

    Args:
        data (dict): refund request fields as defined in spec

    Returns:
        (dict): validated data
    """
    return _validate_request(RefundRequestSchema, data)


def validate_refund_response(data: dict) -> ValidateResponseReturnType:
    """Validate Refund response data.

    Args:
        data (dict): refund response fields as defined in spec

    Returns:
        (dict): validated data
    """
    return _validate_response(RefundResponseSchema, data)


def format_parameter_errors(errors: dict[str, list[str]] = None, data: dict = None) -> ParameterErrorsType:
    """Format parameter errors.

    Args:
        errors (dict): field as a key providing a list of messages
        data (dict): request data

    Returns:
        (list): parameter errors
    """
    errors = errors or {}
    data = data or {}
    parameter_errors: ParameterErrorsType = []
    for field in errors:
        for message in errors[field]:
            error: ParameterErrorType = dict(name=field, message=message)
            if field in data:
                error["value"] = data[field]
            parameter_errors.append(error)

    return parameter_errors


def format_error_response(
    code: ResultCode,
    errors: dict[str, list[str]] = None,
    data: dict = None,
) -> ErrorResponseType:
    r"""Format error response.

    Args:
        code (ResultCode): error code
        errors (dict): field as a key providing a list of messages
        data (dict): request data

    Returns:
        (dict): error response

    Example:
    ```
    {
      "result": {
          "code": "200.300.404",
          "description": "invalid or missing parameter",
          "parameterErrors": [
              {
                  "value": "10.0",
                  "name": "amount",
                  "message": "Must match ^[0-9]{1,8}(\\.[0-9]{2})?$"
              },
              {
                  "value": "ZA",
                  "name": "currency",
                  "message": "Must be a valid ISO-4217, 3-character currency."
              }
          ]
      },
      "timestamp": "2021-08-03T16:16:30.992618Z"
    }
    ```
    """
    errors = errors or {}
    data = data or {}

    result: ErrorResultType = dict(
        code=code.code,
        description=code.description,
    )
    parameter_errors: ParameterErrorsType = format_parameter_errors(errors, data)
    if parameter_errors:
        result["parameterErrors"] = parameter_errors

    response: ErrorResponseType = dict(
        result=result,
        timestamp=datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
    )
    return response
